var TokenSymbol;
(function (TokenSymbol) {
    TokenSymbol[TokenSymbol["DIGIT"] = 1] = "DIGIT";
    TokenSymbol[TokenSymbol["OPEN_PAREN"] = 2] = "OPEN_PAREN";
    TokenSymbol[TokenSymbol["CLOSE_PAREN"] = 3] = "CLOSE_PAREN";
    TokenSymbol[TokenSymbol["PLUS"] = 4] = "PLUS";
    TokenSymbol[TokenSymbol["MINUS"] = 5] = "MINUS";
    TokenSymbol[TokenSymbol["MULTI"] = 6] = "MULTI";
    TokenSymbol[TokenSymbol["DIVIDE"] = 7] = "DIVIDE";
    TokenSymbol[TokenSymbol["END"] = 8] = "END";
    TokenSymbol[TokenSymbol["A"] = -1] = "A";
    TokenSymbol[TokenSymbol["B"] = -2] = "B";
    TokenSymbol[TokenSymbol["C"] = -3] = "C";
    TokenSymbol[TokenSymbol["D"] = -4] = "D";
    TokenSymbol[TokenSymbol["E"] = -5] = "E";
})(TokenSymbol || (TokenSymbol = {}));
const ll1IdxMap = Object.freeze({
    '0': TokenSymbol.DIGIT,
    '1': TokenSymbol.DIGIT,
    '2': TokenSymbol.DIGIT,
    '3': TokenSymbol.DIGIT,
    '4': TokenSymbol.DIGIT,
    '5': TokenSymbol.DIGIT,
    '6': TokenSymbol.DIGIT,
    '7': TokenSymbol.DIGIT,
    '8': TokenSymbol.DIGIT,
    '9': TokenSymbol.DIGIT,
    '(': TokenSymbol.OPEN_PAREN,
    ')': TokenSymbol.CLOSE_PAREN,
    '+': TokenSymbol.PLUS,
    '-': TokenSymbol.MINUS,
    '*': TokenSymbol.MULTI,
    '/': TokenSymbol.DIVIDE,
    $: TokenSymbol.END,
    A: TokenSymbol.A,
    B: TokenSymbol.B,
    C: TokenSymbol.C,
    D: TokenSymbol.D,
    E: TokenSymbol.E,
});
const idx = (c) => ll1IdxMap[c];
const sddTable = [
    'BD',
    '+BD',
    '-BD',
    'CE',
    '0',
    '(A)',
    '+BD',
    '-BD',
    '',
    '*CE',
    '/CE',
    '',
].map(x => x.split('').map(idx));
const MAX_TOKENS = 5;
const MAX_SYMBOLS = 8;
const ll1Table = new Array(MAX_TOKENS + 1);
{
    for (let i = 0; i <= MAX_TOKENS; ++i) {
        ll1Table[i] = new Int8Array(MAX_SYMBOLS + 1).fill(-1);
    }
    ll1Table[1][1] = 0;
    ll1Table[1][2] = 0;
    ll1Table[1][4] = 1;
    ll1Table[1][5] = 2;
    ll1Table[2][1] = 3;
    ll1Table[2][2] = 3;
    ll1Table[3][1] = 4;
    ll1Table[3][2] = 5;
    ll1Table[4][4] = 6;
    ll1Table[4][5] = 7;
    ll1Table[4][3] = 8;
    ll1Table[4][8] = 8;
    ll1Table[5][6] = 9;
    ll1Table[5][7] = 10;
    ll1Table[5][3] = 11;
    ll1Table[5][4] = 11;
    ll1Table[5][5] = 11;
    ll1Table[5][8] = 11;
}

function parseInteger(s, start) {
    let result = 0;
    let i = start;
    for (; i < s.length; ++i) {
        const c = s[i];
        if (!/\d/.test(c))
            break;
        result = result * 10 + Number(c);
    }
    return [i, result];
}
function calculate(rawExpression, parseNumber = parseInteger) {
    const expression = rawExpression.replace(/[\s]+/g, '');
    let cur = 0;
    const result = dfs(idx('A'), 0, 0);
    return cur === expression.length ? result : Number.NaN;
    function dfs(id, syn, inh) {
        if (cur === expression.length) {
            if (id === TokenSymbol.D || id === TokenSymbol.E)
                return inh;
            return Number.NaN;
        }
        const id0 = idx(expression[cur]);
        if (id0 === undefined)
            return Number.NaN;
        if (id === id0) {
            if (id0 === TokenSymbol.DIGIT) {
                const [nextCur, value] = parseNumber(expression, cur);
                if (cur === nextCur)
                    return Number.NaN;
                cur = nextCur;
                return value;
            }
            cur += 1;
            return syn;
        }
        if (id > 0)
            return Number.NaN;
        const ssdId = ll1Table[-id][id0];
        if (ssdId < 0)
            return Number.NaN;
        const tokens = sddTable[ssdId];
        const syn0 = tokens.length > 0 ? dfs(tokens[0], 0, 0) : 0;
        if (Number.isNaN(syn0))
            return Number.NaN;
        switch (ssdId) {
            case 0:
                return dfs(tokens[1], 0, syn0);
            case 1: {
                const val1 = dfs(tokens[1], 0, 0);
                return dfs(tokens[2], 0, val1);
            }
            case 2: {
                const val1 = dfs(tokens[1], 0, 0);
                return dfs(tokens[2], 0, -val1);
            }
            case 3:
                return dfs(tokens[1], 0, syn0);
            case 4:
                return syn0;
            case 5: {
                const result = dfs(tokens[1], 0, 0);
                const result2 = dfs(tokens[2], 0, 0);
                return Number.isNaN(result2) ? Number.NaN : result;
            }
            case 6: {
                const val1 = dfs(tokens[1], 0, 0);
                return dfs(tokens[2], 0, inh + val1);
            }
            case 7: {
                const val1 = dfs(tokens[1], 0, 0);
                return dfs(tokens[2], 0, inh - val1);
            }
            case 8:
                return inh;
            case 9: {
                const val1 = dfs(tokens[1], 0, 0);
                return dfs(tokens[2], 0, inh * val1);
            }
            case 10: {
                const val1 = dfs(tokens[1], 0, 0);
                return dfs(tokens[2], 0, inh / val1);
            }
            case 11:
                return inh;
            default:
                return Number.NaN;
        }
    }
}

export { MAX_SYMBOLS, MAX_TOKENS, TokenSymbol, calculate, calculate as default, idx, ll1IdxMap, ll1Table, parseInteger, sddTable };
